import numpy as np
from numpy.lib.stride_tricks import as_strided

def rolling_view(x, window):
    """
    Returns rolling view (no extra memory needed) over first axis of x.
    Output shape: (x.shape[0]-window+1, window, *x.shape[1:])
    """
    stride = x.strides[0]
    shape = [x.shape[0] - window + 1, window]
    strides = [stride, stride]
    if x.ndim > 1:
        shape += [*x.shape[1:]]
        strides += [*x.strides[1:]]    
    return as_strided(x, shape=shape, strides=strides)


try:
    import statsmodels.stats.api as sms
except:
    pass

def t_interval(x, alpha=0.05, alternative="two-sided"):
    return sms.DescrStatsW(x).tconfint_mean(alpha=alpha, alternative=alternative)


"""
2-sample permutation test, with/without pairing (including block pairing), for the difference and the ratio.

Only implemented for the mean and median. Check functions starting with underscore to implement other statistics.
Will be refined in the future.
"""
import numpy as np
from numba import njit


def _permutation_test_2sample_paired(X1, X2, stat_func, ratio=False, N=int(1e6)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test. Numba won't compile. This is a sketch such that the only thing to implement is stat_func.
    
    Attrs:
            alternative: 
                                           - greater:    H0: mu_1 / mu_2 < observed,       H1: mu_1 / mu_2  >= observed  
                         ratio             - less:       H0: mu_1 / mu_2 > observed,       H1: mu_1 / mu_2  <= observed  
                                           - two-sided:  H0: mu_1 / mu_2 = observed,       H1: mu_1 / mu_2  != observed 
           
                                           - greater:    H0: mu_1 - mu_2 < observed,       H1: mu_1 - mu_2  >= observed  
                         not ratio         - less:       H0: mu_1 - mu_2 > observed,       H1: mu_1 - mu_2  <= observed  
                                           - two-sided:  H0: mu_1 - mu_2 = observed,       H1: mu_1 - mu_2  != observed 
    Returns: p-value
    """
    if ratio:
        def aux(X_paired):
            return stat_func(X_paired[0]) / stat_func(X_paired[1])
    else:
        def aux(X_paired):
            return stat_func(X_paired[0]) - stat_func(X_paired[1])
        
    n = X1.size
    X_paired = np.vstack((X1, X2))
    stat_0 = aux(X_paired)
    
    perm_sample = np.empty((N))
    np.random.seed(seed)
    for i in range(N):
        shuffle = np.random.randint(0, 2, size=n) == 1
        X_shifted = X_paired[:, shuffle][::-1]
        X_still = X_paired[:, ~shuffle]
        X_perm = np.hstack((X_shifted, X_still))
        perm_sample[i] = aux(X_perm)
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")
        
def _permutation_test_2sample_paired_block(X, Y, stat_func, ratio=False, N=int(1e6)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Permutation test for paired blocks. Permutations occur only between blocks: X1(block i) <-> X2(block i). 
    Numba won't compile. This is a sketch such that the only thing to implement is stat_func.
    
    Attrs:
            X, Y:                          ragged arrays or tuples. Each element is an array containing the results for a block. 
            alternative: 
                                           - greater:    H0: mu_1 / mu_2 < observed,       H1: mu_1 / mu_2  >= observed  
                         ratio             - less:       H0: mu_1 / mu_2 > observed,       H1: mu_1 / mu_2  <= observed  
                                           - two-sided:  H0: mu_1 / mu_2 = observed,       H1: mu_1 / mu_2  != observed 
           
                                           - greater:    H0: mu_1 - mu_2 < observed,       H1: mu_1 - mu_2  >= observed  
                         not ratio         - less:       H0: mu_1 - mu_2 > observed,       H1: mu_1 - mu_2  <= observed  
                                           - two-sided:  H0: mu_1 - mu_2 = observed,       H1: mu_1 - mu_2  != observed 
    Returns: p-value
    """
    if ratio:
        def aux(X1, X2):
            return stat_func(X1) / stat_func(X2)
    else:
        def aux(X1, X2):
            return stat_func(X1) - stat_func(X2)
        
    def stack(arr_list):
        return np.array([a for arr in arr_list for a in arr])
        
    stat_0 = aux(stack(X), stack(Y))    
    perm_sample = np.empty((N))
    np.random.seed(seed)
    for i in range(N):
        Xi = []
        Yi = []
        for xi, yi in zip(X, Y):
            z = np.hstack((xi, yi))
            np.random.shuffle(z)
            Xi.append(z[:xi.size])
            Yi.append(z[xi.size:])
        perm_sample[i] = aux(stack(Xi), stack(Yi))
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")

def _permutation_test_2sample_not_paired(X1, X2, stat_func, ratio=False, N=int(1e6) - 1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Non-paired permutation test. Numba won't compile. This is a sketch such that the only thing to implement is stat_func.
    
    Attrs:
            alternative: 
                                           - greater:    H0: mu_1 / mu_2 < observed,       H1: mu_1 / mu_2  >= observed  
                         ratio             - less:       H0: mu_1 / mu_2 > observed,       H1: mu_1 / mu_2  <= observed  
                                           - two-sided:  H0: mu_1 / mu_2 = observed,       H1: mu_1 / mu_2  != observed 
           
                                           - greater:    H0: mu_1 - mu_2 < observed,       H1: mu_1 - mu_2  >= observed  
                         not ratio         - less:       H0: mu_1 - mu_2 > observed,       H1: mu_1 - mu_2  <= observed  
                                           - two-sided:  H0: mu_1 - mu_2 = observed,       H1: mu_1 - mu_2  != observed 
    Returns: p-value
    """
    n1 = X1.size
    if ratio:
        def aux(X, n1): # pass n1 to avoid numba error.
            return stat_func(X[:n1]) / stat_func(X[n1:])
    else:
        def aux(X, n1):
            return stat_func(X[:n1]) - stat_func(X[n1:])
    X = np.hstack((X1, X2))
    stat_0 = aux(X)
       
    perm_sample = np.empty((N)) # permutation distribution
    np.random.seed(seed)
    for i in range(N):       
        np.random.shuffle(X)
        perm_sample[i] = aux(X, n1)
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")

@njit
def permutation_test_2sample_paired_median(X1, X2, ratio=False, N=int(1e6)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test for the median.
    
    Attrs:
            alternative: 
                                           - greater:    H0: me_1 / me_2 < observed,       H1: me_1 / me_2  >= observed  
                         ratio             - less:       H0: me_1 / me_2 > observed,       H1: me_1 / me_2  <= observed  
                                           - two-sided:  H0: me_1 / me_2 = observed,       H1: me_1 / me_2  != observed 
           
                                           - greater:    H0: me_1 - me_2 < observed,       H1: me_1 - me_2  >= observed  
                         not ratio         - less:       H0: me_1 - me_2 > observed,       H1: me_1 - me_2  <= observed  
                                           - two-sided:  H0: me_1 - me_2 = observed,       H1: me_1 - me_2  != observed 
            
    Returns: p-value
    """
    if ratio:
        def aux(X_paired):
            return np.median(X_paired[0]) / np.median(X_paired[1])
    else:
        def aux(X_paired):
            return np.median(X_paired[0]) - np.median(X_paired[1])
        
    n = X1.size
    X_paired = np.vstack((X1, X2))
    stat_0 = aux(X_paired)
    
    perm_sample = np.empty((N))
    np.random.seed(seed)
    for i in range(N):
        shuffle = np.random.randint(0, 2, size=n) == 1
        X_shifted = X_paired[:, shuffle][::-1]
        X_still = X_paired[:, ~shuffle]
        X_perm = np.hstack((X_shifted, X_still))
        perm_sample[i] = aux(X_perm)
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")

        
@njit
def permutation_test_2sample_paired_block_median(X, Y, ratio=False, N=int(1e6)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Permutation test for the mean for paired blocks. Permutations occur only between blocks: X1(block i) <-> X2(block i). 
    Numba won't compile. This is a sketch such that the only thing to implement is stat_func.
    
    Attrs:
            X, Y:                          ragged arrays or tuples. Each element is an array containing the results for a block. 
            alternative: 
                                           - greater:    H0: mu_1 / mu_2 < observed,       H1: mu_1 / mu_2  >= observed  
                         ratio             - less:       H0: mu_1 / mu_2 > observed,       H1: mu_1 / mu_2  <= observed  
                                           - two-sided:  H0: mu_1 / mu_2 = observed,       H1: mu_1 / mu_2  != observed 
           
                                           - greater:    H0: mu_1 - mu_2 < observed,       H1: mu_1 - mu_2  >= observed  
                         not ratio         - less:       H0: mu_1 - mu_2 > observed,       H1: mu_1 - mu_2  <= observed  
                                           - two-sided:  H0: mu_1 - mu_2 = observed,       H1: mu_1 - mu_2  != observed 
    Returns: p-value
    """
    if ratio:
        def aux(X1, X2):
            return np.median(X1) / np.median(X2)
    else:
        def aux(X1, X2):
            return np.median(X1) - np.median(X2)
        
    def stack(arr_list):
        return np.array([a for arr in arr_list for a in arr])
        
    stat_0 = aux(stack(X), stack(Y))    
    perm_sample = np.empty((N))
    np.random.seed(seed)
    for i in range(N):
        Xi = []
        Yi = []
        for xi, yi in zip(X, Y):
            z = np.hstack((xi, yi))
            np.random.shuffle(z)
            Xi.append(z[:xi.size])
            Yi.append(z[xi.size:])
        perm_sample[i] = aux(stack(Xi), stack(Yi))
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")
        
        
@njit    
def permutation_test_2sample_not_paired_median(X1, X2, ratio=False, N=int(1e6) - 1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Non-paired permutation test for the median.
    
   Attrs:
            alternative: 
                                           - greater:    H0: me_1 / me_2 < observed,       H1: me_1 / me_2  >= observed  
                         ratio             - less:       H0: me_1 / me_2 > observed,       H1: me_1 / me_2  <= observed  
                                           - two-sided:  H0: me_1 / me_2 = observed,       H1: me_1 / me_2  != observed 
           
                                           - greater:    H0: me_1 - me_2 < observed,       H1: me_1 - me_2  >= observed  
                         not ratio         - less:       H0: me_1 - me_2 > observed,       H1: me_1 - me_2  <= observed  
                                           - two-sided:  H0: me_1 - me_2 = observed,       H1: me_1 - me_2  != observed 
    Returns: p-value
    """
    n1 = X1.size
    if ratio:
        def aux(X, n1): # pass n1 to avoid numba error.
            return np.median(X[:n1]) / np.median(X[n1:])
    else:
        def aux(X, n1):
            return np.median(X[:n1]) - np.median(X[n1:])
    X = np.hstack((X1, X2))
    stat_0 = aux(X, n1)
       
    perm_sample = np.empty((N)) # permutation distribution
    np.random.seed(seed)
    for i in range(N):       
        np.random.shuffle(X)
        perm_sample[i] = aux(X, n1)
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")
        
@njit
def permutation_test_2sample_paired_mean(X1, X2, ratio=False, N=int(1e6)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test for the mean
    
    Attrs:
            alternative: 
                                           - greater:    H0: mu_1 / mu_2 < observed,       H1: mu_1 / mu_2  >= observed  
                         ratio             - less:       H0: mu_1 / mu_2 > observed,       H1: mu_1 / mu_2  <= observed  
                                           - two-sided:  H0: mu_1 / mu_2 = observed,       H1: mu_1 / mu_2  != observed 
           
                                           - greater:    H0: mu_1 - mu_2 < observed,       H1: mu_1 - mu_2  >= observed  
                         not ratio         - less:       H0: mu_1 - mu_2 > observed,       H1: mu_1 - mu_2  <= observed  
                                           - two-sided:  H0: mu_1 - mu_2 = observed,       H1: mu_1 - mu_2  != observed 
    Returns: p-value
    """
    if ratio:
        def aux(X_paired):
            return np.mean(X_paired[0]) / np.mean(X_paired[1])
    else:
        def aux(X_paired):
            return np.mean(X_paired[0]) - np.mean(X_paired[1])
        
    n = X1.size
    X_paired = np.vstack((X1, X2))
    stat_0 = aux(X_paired)
    
    perm_sample = np.empty((N))
    np.random.seed(seed)
    for i in range(N):
        shuffle = np.random.randint(0, 2, size=n) == 1
        X_shifted = X_paired[:, shuffle][::-1]
        X_still = X_paired[:, ~shuffle]
        X_perm = np.hstack((X_shifted, X_still))
        perm_sample[i] = aux(X_perm)
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")

@njit
def permutation_test_2sample_paired_block_mean(X, Y, ratio=False, N=int(1e6)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Permutation test for the mean for paired blocks. Permutations occur only between blocks: X1(block i) <-> X2(block i). 
    Numba won't compile. This is a sketch such that the only thing to implement is stat_func.
    
    Attrs:
            X, Y:                          ragged arrays or tuples. Each element is an array containing the results for a block. 
            alternative: 
                                           - greater:    H0: mu_1 / mu_2 < observed,       H1: mu_1 / mu_2  >= observed  
                         ratio             - less:       H0: mu_1 / mu_2 > observed,       H1: mu_1 / mu_2  <= observed  
                                           - two-sided:  H0: mu_1 / mu_2 = observed,       H1: mu_1 / mu_2  != observed 
           
                                           - greater:    H0: mu_1 - mu_2 < observed,       H1: mu_1 - mu_2  >= observed  
                         not ratio         - less:       H0: mu_1 - mu_2 > observed,       H1: mu_1 - mu_2  <= observed  
                                           - two-sided:  H0: mu_1 - mu_2 = observed,       H1: mu_1 - mu_2  != observed 
    Returns: p-value
    """
    if ratio:
        def aux(X1, X2):
            return np.mean(X1) / np.mean(X2)
    else:
        def aux(X1, X2):
            return np.mean(X1) - np.mean(X2)
        
    def stack(arr_list):
        return np.array([a for arr in arr_list for a in arr])
        
    stat_0 = aux(stack(X), stack(Y))    
    perm_sample = np.empty((N))
    np.random.seed(seed)
    for i in range(N):
        Xi = []
        Yi = []
        for xi, yi in zip(X, Y):
            z = np.hstack((xi, yi))
            np.random.shuffle(z)
            Xi.append(z[:xi.size])
            Yi.append(z[xi.size:])
        perm_sample[i] = aux(stack(Xi), stack(Yi))
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")
        
@njit
def permutation_test_2sample_not_paired_mean(X1, X2, ratio=False, N=int(1e6) - 1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Permutation test for the differences or ratio of the means (non-paired).
    
    Attrs:
            alternative:  - greater:    H0: mu_1 - mu_2 < observed,       H1: mu_1 - mu_2  >= observed
                          - less:       H0: mu_1 - mu_2 > observed,       H1: mu_1 - mu_2  <= observed
                          - two-sided:  H0: mu_1 - mu_2 = observed,       H1: mu_1 - mu_2  != observed
    Returns: p-value
    """
    n1 = X1.size
    if ratio:
        def aux(X, n1): # pass n1 to avoid numba error.
            return X[:n1].mean() / X[n1:].mean()
    else:
        def aux(X, n1):
            return X[:n1].mean() - X[n1:].mean()
    X = np.hstack((X1, X2))
    stat_0 = aux(X, n1)
       
    perm_sample = np.empty((N)) # permutation distribution
    np.random.seed(seed)
    for i in range(N):       
        np.random.shuffle(X)
        perm_sample[i] = aux(X, n1)
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")

@njit
def permutation_test_2sample_paired_diffmean(X1, X2, N=int(1e6)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test optimized for the mean of the differences. about 2x faster than the default _permutation_test_2sample_paired with stat_func = np.mean.
    Coincides with the test for the differences in means. (Not as the median)
    Attrs:
            alternative:  - greater:    H0: mu(X1-X2) < <X1-X2>_sample,       H1: mu(X1-X2) >= <X1-X2>_sample
                          - less:       H0: mu(X1-X2) > <X1-X2>_sample,       H1: mu(X1-X2) <= <X1-X2>_sample
                          - two-sided:  H0: mu(X1-X2) = <X1-X2>_sample,       H1: mu(X1-X2) != <X1-X2>_sample
    Returns: p-value
    """
    dX = X1 - X2
    n = dX.size
    mean_0 = dX.mean()
    
    perm_sample = np.empty((N))
    np.random.seed(seed)
    for i in range(N):        
        shuffle = np.random.randint(0, 2, size=n) == 1
        dX_perm = dX.copy()
        dX_perm[shuffle] *= -1
        perm_sample[i] = (dX_perm).mean()
    
    if alternative == "greater":
        return (1 + (perm_sample >= mean_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= mean_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= mean_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= mean_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")
        
@njit
def permutation_test_2sample_paired_diffmedian(X1, X2, N=int(1e6)-1, alternative="two-sided", tolerance=1.5e-8, seed=0):
    """
    Paired permutation test optimized for the median of the differences. about 2x faster than the default _permutation_test_2sample_paired with stat_func = np.mean.
    NOTE: Does not coincide with the test for the differences in medians.
    
    Attrs:
            alternative:  - greater:    H0: me(X1-X2) < observed,       H1: me(X1-X2) >= observed
                          - less:       H0: me(X1-X2) > observed,       H1: me(X1-X2) <= observed
                          - two-sided:  H0: me(X1-X2) = observed,       H1: me(X1-X2) != observed.
    Returns: p-value
    """
    dX = X1 - X2
    n = dX.size
    stat_0 = np.median(dX)
    
    perm_sample = np.empty((N))
    np.random.seed(seed)
    for i in range(N):        
        shuffle = np.random.randint(0, 2, size=n) == 1
        dX_perm = dX.copy()
        dX_perm[shuffle] *= -1
        perm_sample[i] = np.median(dX_perm)
    
    if alternative == "greater":
        return (1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1)
    elif alternative == "less":
        return (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
    elif alternative == "two-sided":
        return 2 * np.fmin((1 + (perm_sample >= stat_0 - tolerance).sum()) / (N + 1),
                           (1 + (perm_sample <= stat_0 + tolerance).sum()) / (N + 1)
                          )
    else:
        raise ValueError("alternative not valid")


"""
Helper funcs for plotly figures
"""
import numpy as np
import pandas as pd
import warnings
try:
    import plotly.express as px
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
except:
    warnings.warn("'plotly' not available.", RuntimeWarning)
from collections import defaultdict
from collections.abc import Iterable
from functools import partial
from .. import _helper
from .base import color_std, plotly_default_colors

def get_common_range(fig, axes=["x", "y"], offset_mpl=[0,0], offset_constant=[0,0]):
    data = defaultdict(list)
    for plot in fig.data:
        for ax, off_m, off_c in zip(axes, offset_mpl, offset_constant):          
            if hasattr(plot, f"error_{ax}") and getattr(plot, f"error_{ax}").array is not None:
                additions = [np.array([*plot[f"error_{ax}"]["array"]]), -np.array([*plot[f"error_{ax}"]["array"]])] 
            else:
                additions = [0]
            for addition in additions:
                arr = (plot[ax] + addition)[~np.isnan(plot[ax])]
                arr_min, arr_max = arr.min(), arr.max()
                data[f"{ax}-min"].append(arr_min * (1-off_m if arr_min > 0 else 1+off_m) - off_c)
                data[f"{ax}-max"].append(arr_max * (1+off_m if arr_max > 0 else 1-off_m) + off_c)
    for k, v in data.items():
        func = min if "min" in k else max
        data[k] = func(v)
    ranges = {ax: [data[f"{ax}-min"], data[f"{ax}-max"]] for ax in axes}
    return ranges

def get_nplots(fig):
    return sum(1 for x in fig.layout if "xaxis" in x)

def mod_delete_axes(fig, axes=["x", "y"]):
    non_visible_axes_specs = dict(visible=False, showgrid=False, zeroline=False) 
    return {f"{ax}axis{i}": non_visible_axes_specs for ax in axes for i in [""] + [*range(1, get_nplots(fig) + 1)]}

def get_mod_layout(key, val=None):
    def mod_layout(fig, val, axes=["x","y"]):
        if isinstance(val, Iterable) and not isinstance(val, str):
            return {"{}axis{}_{}".format(ax, i, key): v for (ax, v) in zip(axes, val) for i in [""] + [*range(1, get_nplots(fig) + 1)]}
        else:
            return {"{}axis{}_{}".format(ax, i, key): val for ax in axes for i in [""] + [*range(1, get_nplots(fig) + 1)]}
    if val is None:
        return mod_layout
    else:
        def mod_layout_fixed_val(fig, axes=["x", "y"]):
            return mod_layout(fig, val, axes)
        return mod_layout_fixed_val

mod_dashes           = partial(_helper.sequence_or_stream, ["solid", "dash", "dot"])
mod_ticksize         = get_mod_layout("tickfont_size")
mod_logaxes          = get_mod_layout("type", "log") 
mod_expfmt           = get_mod_layout("exponentformat", "power")
mod_range            = get_mod_layout("range")
mod_logaxes_expfmt   = lambda fig, axes=["x", "y"]: {**mod_logaxes(fig, axes=axes), **mod_expfmt(fig, axes=axes)}

def mod_common_range(fig, axes=["x", "y"], **kwargs):
    return mod_range(fig, val=get_common_range(fig, axes=axes, **kwargs), axes=axes)

def get_figure(height=800, width=1000, ticksize=32, font_size=40, margin=None, font_family="sans-serif", hovermode=False, delete_axes=False, **kwargs):
    fig = go.Figure(layout=dict(margin=dict(l=100, r=20, b=80, t=20, pad=1) if margin is None else margin,
                                height=height, width=width, yaxis=dict(tickfont_size=ticksize),
                                xaxis=dict(tickfont_size=ticksize), font_size=font_size, legend_font_size=font_size,
                                font_family=font_family, hovermode=hovermode,
                                **kwargs))
    if delete_axes:
        fig.update_layout(**mod_delete_axes(fig), margin=dict(l=0, t=0, b=0, r=0), paper_bgcolor="rgba(0,0,0,0)", plot_bgcolor="rgba(0,0,0,0)")
    return fig

def get_subplots(cols, rows=1, horizontal_spacing=0.03, vertical_spacing=0.03, height=None, width=2500, ticksize=32, font_size=40, font_family="sans-serif",
                 hovermode=False, delete_axes=False, shared_xaxes=True, shared_yaxes=True, layout_kwargs={}, 
                 **make_subplots_kwargs):
    height = 800*rows if height is None else height
    fig = make_subplots(figure=go.Figure(layout=dict(margin=dict(l=100, r=20, b=80, t=60, pad=1), height=height, width=width)),
                        shared_yaxes=shared_yaxes, shared_xaxes=shared_xaxes,                        
                        horizontal_spacing=horizontal_spacing, vertical_spacing=vertical_spacing, rows=rows, cols=cols,
                        **make_subplots_kwargs
                       )
                    
    fig.for_each_annotation(lambda a: a.update(font={'size':font_size, "family":font_family}))
    fig.update_layout(**mod_ticksize(fig, val=ticksize), legend_font_size=font_size, hovermode=hovermode, **layout_kwargs)
    if delete_axes:
        fig.update_layout(**mod_delete_axes(fig), margin=dict(l=0, t=0, b=0, r=0), paper_bgcolor="rgba(0,0,0,0)", plot_bgcolor="rgba(0,0,0,0)")
    return fig

def transparent_colorscale(fig, threshold=1e-10):
    """Values below threshold are invisible."""
    colorscale = fig.layout["coloraxis"]["colorscale"]
    low_limit = colorscale[0]
    new_low_limit = (threshold, low_limit[1])
    new_colorscale = ((0, 'rgba(0,0,0,0)'), new_low_limit, *colorscale[1:])
    return new_colorscale

def multiindex_to_label(i, depth=2):
    return [i.get_level_values(k).to_list() for k in range(depth)]

def set_multicategory_from_df(fig, df):
    fig.update_layout(xaxis_type="multicategory", yaxis_type="multicategory")
    fig.data[0]["x"] = multiindex_to_label(df.columns)
    fig.data[0]["y"] = multiindex_to_label(df.index)
    return

def violin(df, CI=None, CI_line="mean", CI_width=0.05):
    """
    Violin plot including optionally the CI.
    
    Attributes:
        -df:   melted DataFrame. Contains only two columns: variable name (x) and value (y).
                                 The column names set the OX and OY labels.
    """
    x, y = df.columns
    fig = get_figure(xaxis_title=x, yaxis_title=y)
    fig.add_trace(go.Violin(x=df[x], y=df[y], showlegend=False))
    if CI is not None:
        CI_color = color_std(plotly_default_colors(maxlen=2)[-1], opacity=0.2)
        CI_stats = getattr(df.groupby(x), CI_line)().values.squeeze()
        for i, (ci, ci_stat) in enumerate(zip(CI, CI_stats)):
            fig.add_trace(go.Scatter(x=[f"Trial {i+2}"]*2, y=ci, mode="markers",
                                     marker=dict(color=CI_color, symbol=["arrow-bar-up", "arrow-bar-down"], size=8, line=dict(color="gray", width=2))
                                     , showlegend=False))
            fig.add_shape(type="rect", xref="x", yref="y", line=dict(color="gray",width=3), fillcolor=CI_color, x0=i-CI_width, y0=ci[0], x1=i+CI_width, y1=ci[1])
            fig.add_shape(type="line", xref="x", yref="y", line=dict(color="gray", width=4),  x0=i-CI_width, y0=ci_stat, x1=i+CI_width, y1=ci_stat)
    return fig


"""
pandas utils

Includes    latex_table:      f:  pd.DataFrame -> latex table.
"""
import pandas as pd
import numpy as np
from functools import reduce     

def latex_table(df, index=False, **kwargs):
    """Pandas DataFrame -> Latex table."""
    col_format = "c" if isinstance(df, pd.core.series.Series) else "c"*len(df.columns)
    if index:
        col_format += "c"
    table_replacements = (("\\toprule", "\\toprule "*2),
                          ("\\bottomrule", "\\bottomrule "*2)
    )
    text_replacements = (("\\textbackslash ", "\\"),
                         ("\{", "{"), 
                         ("\}", "}"),
                         ("\$", "$"),
                         ("\_", "_"),
                         ("\\textasciicircum ", "^")
    )
    table_formatter = lambda x:  reduce(lambda a, kv: a.replace(*kv), table_replacements, x)
    text_formatter = lambda x: reduce(lambda a, kv: a.replace(*kv), text_replacements, x)
    formatter = lambda x: text_formatter(table_formatter(x))
    print(formatter(df.to_latex(index=index, column_format=col_format, **kwargs)))
    return

def expand_sequences(df, dt=1, maxlen=None):
    """
    Input: DataFrame. Each element is an array and all arrays start at the same time and have the same time step dt.
    Returns: MultiColumn DataFrame: (df.index,  (df.columns, time_steps))
    """
    if df.isna().values.any():
        if maxlen is None:
            maxlen = int(df.applymap(lambda x: x.size if isinstance(x, np.ndarray) else np.NaN).max().max())
        df_padded = df.applymap(lambda x: np.hstack((x, np.NaN*np.empty((maxlen-x.size)))) if isinstance(x, np.ndarray) else np.NaN*np.empty((maxlen)))
    else:
        if maxlen is None:
            maxlen = int(df.applymap(lambda x: x.size).values.max())
        df_padded = df.applymap(lambda x: np.hstack((x, np.NaN*np.empty((maxlen-x.size)))))
    df_padded_arr = np.stack([np.vstack(x) for x in df_padded.values]) # shape (df.shape[0], df.shape[1], time_steps)
    return pd.DataFrame(df_padded_arr.reshape((df.shape[0], -1)), 
                        index = df.index, 
                        columns = pd.MultiIndex.from_product([df.columns, dt*np.arange(maxlen)]))



from tidypath import savedata, savefig, add_arg, modify_arg, delete_arg
from . import np_utils
from . import pd_utils
from . import storage
from . import plots
from . import stats
from .stats.rtopy import resample
from .stats.test import permutation
